# Copyright  2016-2022 Maël Azimi <m.a@moul.re>
#
# Silkaj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Silkaj is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Silkaj. If not, see <https://www.gnu.org/licenses/>.


import sys
from pathlib import Path
from typing import Dict

import click
from duniterpy.api import bma
from duniterpy.documents.block_id import BlockID
from duniterpy.documents.document import MalformedDocumentError
from duniterpy.documents.identity import Identity
from duniterpy.documents.revocation import Revocation
from duniterpy.key.verifying_key import VerifyingKey

from silkaj import auth, network, tui
from silkaj.blockchain.tools import get_currency
from silkaj.constants import FAILURE_EXIT_STATUS, SUCCESS_EXIT_STATUS
from silkaj.public_key import gen_pubkey_checksum
from silkaj.wot import idty_tools
from silkaj.wot import tools as w_tools

REVOCATION_LOCAL_PATH = "revocation.txt"


@click.command(
    "save",
    help="Create and save a revocation document. Optionnaly takes the document filename.",
)
@click.argument(
    "file",
    default=REVOCATION_LOCAL_PATH,
)
@click.pass_context
def save(ctx: click.Context, file: str) -> None:
    currency = get_currency()

    key = auth.auth_method()
    gen_pubkey_checksum(key.pubkey)
    _id = (w_tools.choose_identity(key.pubkey))[0]
    rev_doc = create_revocation_doc(_id, key.pubkey, currency)
    rev_doc.sign(key)

    if ctx.obj["DRY_RUN"]:
        click.echo(rev_doc.signed_raw())
        return

    idty_table = idty_tools.display_identity(rev_doc.identity)
    click.echo(idty_table.draw())
    if ctx.obj["DISPLAY_DOCUMENT"]:
        click.echo(rev_doc.signed_raw())

    confirm_message = "Do you want to save the revocation document for this identity?"
    if click.confirm(confirm_message):
        save_doc(file, rev_doc.signed_raw(), key.pubkey)
    else:
        click.echo("Ok, goodbye!")


@click.command(
    "revoke",
    help="Create and publish revocation document. Will revoke the identity immediately.",
)
@click.pass_context
def revoke_now(ctx: click.Context) -> None:
    currency = get_currency()

    warn_before_dry_run_or_display(ctx)

    key = auth.auth_method()
    gen_pubkey_checksum(key.pubkey)
    _id = (w_tools.choose_identity(key.pubkey))[0]
    rev_doc = create_revocation_doc(_id, key.pubkey, currency)
    rev_doc.sign(key)

    if ctx.obj["DRY_RUN"]:
        click.echo(rev_doc.signed_raw())
        return

    idty_table = idty_tools.display_identity(rev_doc.identity)
    click.echo(idty_table.draw())
    if ctx.obj["DISPLAY_DOCUMENT"]:
        click.echo(rev_doc.signed_raw())

    warn_before_sending_document()
    network.send_document(bma.wot.revoke, rev_doc)


@click.command(
    "verify",
    help="Verifies that a revocation document is correctly formatted and matches an \
existing identity.\nOptionnaly takes the document filename.",
)
@click.argument(
    "file",
    default=REVOCATION_LOCAL_PATH,
)
@click.pass_context
def verify(ctx: click.Context, file: str) -> None:
    rev_doc = verify_document(file)

    if ctx.obj["DRY_RUN"]:
        click.echo(rev_doc.signed_raw())
        return

    idty_table = idty_tools.display_identity(rev_doc.identity)
    click.echo(idty_table.draw())
    if ctx.obj["DISPLAY_DOCUMENT"]:
        click.echo(rev_doc.signed_raw())

    click.echo("Revocation document is valid.")


@click.command(
    "publish",
    help="Publish revocation document. Identity will be revoked immediately.\n\
Optionnaly takes the document filename.",
)
@click.argument(
    "file",
    default=REVOCATION_LOCAL_PATH,
)
@click.pass_context
def publish(ctx: click.Context, file: str) -> None:

    warn_before_dry_run_or_display(ctx)

    rev_doc = verify_document(file)
    if ctx.obj["DRY_RUN"]:
        click.echo(rev_doc.signed_raw())
        return

    idty_table = idty_tools.display_identity(rev_doc.identity)
    click.echo(idty_table.draw())
    if ctx.obj["DISPLAY_DOCUMENT"]:
        click.echo(rev_doc.signed_raw())

    warn_before_sending_document()
    network.send_document(bma.wot.revoke, rev_doc)


def warn_before_dry_run_or_display(ctx: click.Context) -> None:
    if ctx.obj["DRY_RUN"]:
        click.echo("WARNING: the document will only be displayed and will not be sent.")


def warn_before_sending_document() -> None:
    click.secho("/!\\WARNING/!\\", blink=True, fg="red")
    click.echo(
        "This identity will be revoked.\n\
It will cease to be member and to create the Universal Dividend.\n\
All currently sent certifications will remain valid until they expire."
    )
    tui.send_doc_confirmation("revocation document immediately")


def create_revocation_doc(_id: Dict, pubkey: str, currency: str) -> Revocation:
    """
    Creates an unsigned revocation document.
    _id is the dict object containing id infos from request wot.requirements
    """
    idty = Identity(
        currency=currency,
        pubkey=pubkey,
        uid=_id["uid"],
        block_id=BlockID.from_str(_id["meta"]["timestamp"]),
    )
    idty.signature = _id["self"]
    return Revocation(
        currency=currency,
        identity=idty,
    )


def save_doc(path: str, content: str, pubkey: str) -> None:
    pubkey_cksum = gen_pubkey_checksum(pubkey)
    rev_path = Path(path)
    # Ask confirmation if the file exists
    if rev_path.is_file():
        if click.confirm(
            f"Would you like to erase existing file `{path}` with the \
generated revocation document corresponding to {pubkey_cksum} public key?"
        ):
            rev_path.unlink()
        else:
            click.echo("Ok, goodbye!")
            sys.exit(SUCCESS_EXIT_STATUS)
    # write doc
    with open(rev_path, "w", encoding="utf-8") as file:
        file.write(content)
    click.echo(
        f"Revocation document file stored into `{path}` for following public key: {pubkey_cksum}"
    )


def verify_document(doc: str) -> Revocation:
    """
    This checks that:
      - that the revocation signature is valid.
      - if the identity is unique (warns the user)
    It returns the revocation document or exits.
    """
    error_invalid_sign = "Error: the signature of the revocation document is invalid."
    error_invalid_doc = (
        f"Error: {doc} is not a revocation document, or is not correctly formatted."
    )

    if not Path(doc).is_file():
        sys.exit(f"Error: file {doc} does not exist")
    with open(doc, encoding="utf-8") as document:
        original_doc = document.read()

    try:
        rev_doc = Revocation.from_signed_raw(original_doc)
    except (MalformedDocumentError, IndexError):
        sys.exit(error_invalid_doc)

    verif_key = VerifyingKey(rev_doc.pubkey)
    if not verif_key.check_signature(rev_doc.raw(), rev_doc.signature):
        sys.exit(error_invalid_sign)

    many_idtys = idty_tools.check_many_identities(rev_doc)
    if many_idtys:
        return rev_doc
    sys.exit(FAILURE_EXIT_STATUS)



# Copyright  2016-2022 Maël Azimi <m.a@moul.re>
#
# Silkaj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Silkaj is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Silkaj. If not, see <https://www.gnu.org/licenses/>.

import functools
from typing import Dict

from duniterpy.api.bma import blockchain

from silkaj.network import client_instance


@functools.lru_cache(maxsize=1)
def get_blockchain_parameters() -> Dict:
    client = client_instance()
    return client(blockchain.parameters)


@functools.lru_cache(maxsize=1)
def get_head_block() -> Dict:
    client = client_instance()
    return client(blockchain.current)


@functools.lru_cache(maxsize=1)
def get_currency() -> str:
    return get_head_block()["currency"]



# Copyright  2016-2022 Maël Azimi <m.a@moul.re>
#
# Silkaj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Silkaj is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Silkaj. If not, see <https://www.gnu.org/licenses/>.

import urllib

import click

from silkaj.network import exit_on_http_error
from silkaj.public_key import gen_pubkey_checksum, is_pubkey_and_check
from silkaj.wot import tools as wt


@click.command("lookup", help="User identifier and public key lookup")
@click.argument("uid_pubkey")
def lookup_cmd(uid_pubkey: str) -> None:
    checked_pubkey = is_pubkey_and_check(uid_pubkey)
    if checked_pubkey:
        uid_pubkey = str(checked_pubkey)

    try:
        lookups = wt.wot_lookup(uid_pubkey)
    except urllib.error.HTTPError as e:
        exit_on_http_error(e, 404, f"No identity found for {uid_pubkey}")

    content = f"Public keys or user id found matching '{uid_pubkey}':\n"
    for lookup in lookups:
        for identity in lookup["uids"]:
            pubkey_checksum = gen_pubkey_checksum(lookup["pubkey"])
            content += f'\n→ {pubkey_checksum} ↔ {identity["uid"]}'
    click.echo(content)



# Copyright  2016-2022 Maël Azimi <m.a@moul.re>
#
# Silkaj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Silkaj is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Silkaj. If not, see <https://www.gnu.org/licenses/>.

import functools
from typing import List, Tuple, Union

from duniterpy.api import bma
from duniterpy.documents.transaction import InputSource, OutputSource

from silkaj.network import client_instance
from silkaj.public_key import gen_pubkey_checksum
from silkaj.wot import tools as wt


def display_amount(
    tx: List, message: str, amount: float, ud_value: float, currency_symbol: str
) -> None:
    """
    Displays an amount in unit and relative reference.
    """
    UD_amount = str(round((amount / ud_value), 2))
    unit_amount = str(amount / 100)
    tx.append(
        [
            f"{message} (unit|relative)",
            f"{unit_amount} {currency_symbol} | {UD_amount} UD {currency_symbol}",
        ]
    )


def display_pubkey(tx: List, message: str, pubkey: str) -> None:
    """
    Displays a pubkey and the eventually associated identity
    """
    tx.append([f"{message} (pubkey:checksum)", gen_pubkey_checksum(pubkey)])
    idty = wt.is_member(pubkey)
    if idty:
        tx.append([f"{message} (id)", idty["uid"]])


def get_amount_from_pubkey(pubkey: str) -> List[int]:
    listinput, amount = get_sources(pubkey)

    totalAmountInput = 0
    for _input in listinput:
        totalAmountInput += amount_in_current_base(_input)
    return [totalAmountInput, amount]


def get_sources(pubkey: str) -> Tuple[List[InputSource], int]:
    client = client_instance()
    # Sources written into the blockchain
    sources = client(bma.tx.sources, pubkey)

    listinput = []
    amount = 0
    for source in sources["sources"]:
        if source["conditions"] == f"SIG({pubkey})":
            listinput.append(
                InputSource(
                    amount=source["amount"],
                    base=source["base"],
                    source=source["type"],
                    origin_id=source["identifier"],
                    index=source["noffset"],
                )
            )
            amount += amount_in_current_base(listinput[-1])

    # pending source
    history = (client(bma.tx.pending, pubkey))["history"]
    pendings = history["sending"] + history["receiving"] + history["pending"]

    # add pending output
    pending_sources = []
    for pending in pendings:
        for i, output in enumerate(pending["outputs"]):
            # duniterpy#80
            outputsplited = output.split(":")
            if outputsplited[2] == f"SIG({pubkey})":
                inputgenerated = InputSource(
                    amount=int(outputsplited[0]),
                    base=int(outputsplited[1]),
                    source="T",
                    origin_id=pending["hash"],
                    index=i,
                )
                if inputgenerated not in listinput:
                    # add pendings before blockchain sources for change txs
                    listinput.insert(0, inputgenerated)

        for _input in pending["inputs"]:
            pending_sources.append(InputSource.from_inline(_input))

    # remove input already used
    for _input in pending_sources:
        if _input in listinput:
            listinput.remove(_input)

    return listinput, amount


@functools.lru_cache(maxsize=1)
def get_ud_value() -> int:
    client = client_instance()
    blockswithud = client(bma.blockchain.ud)
    NBlastUDblock = blockswithud["result"]["blocks"][-1]
    lastUDblock = client(bma.blockchain.block, NBlastUDblock)
    return lastUDblock["dividend"] * 10 ** lastUDblock["unitbase"]


def amount_in_current_base(source: Union[InputSource, OutputSource]) -> int:
    """
    Get amount in current base from input or output source
    """
    return source.amount * 10**source.base



# Copyright  2016-2022 Maël Azimi <m.a@moul.re>
#
# Silkaj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Silkaj is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Silkaj. If not, see <https://www.gnu.org/licenses/>.

import click
from pendulum import from_timestamp

from silkaj.blockchain.tools import get_head_block
from silkaj.constants import ALL
from silkaj.network import determine_endpoint
from silkaj.tools import get_currency_symbol


@click.command("info", help="Display information about currency")
def currency_info() -> None:
    head_block = get_head_block()
    ep = determine_endpoint()
    current_time = from_timestamp(head_block["time"], tz="local")
    mediantime = from_timestamp(head_block["medianTime"], tz="local")
    print(
        "Connected to node:",
        ep.host,
        ep.port,
        "\nCurrent block number:",
        head_block["number"],
        "\nCurrency name:",
        get_currency_symbol(),
        "\nNumber of members:",
        head_block["membersCount"],
        "\nMinimal Proof-of-Work:",
        head_block["powMin"],
        "\nCurrent time:",
        current_time.format(ALL),
        "\nMedian time:",
        mediantime.format(ALL),
        "\nDifference time:",
        current_time.diff_for_humans(mediantime, True),
    )



# Copyright  2016-2022 Maël Azimi <m.a@moul.re>
#
# Silkaj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Silkaj is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Silkaj. If not, see <https://www.gnu.org/licenses/>.

import shutil
import sys
import urllib
from typing import List, Union

import click
import pendulum
from duniterpy.api import bma
from duniterpy.documents import BlockID, Identity, Revocation
from texttable import Texttable

from silkaj.constants import ALL
from silkaj.network import client_instance
from silkaj.public_key import gen_pubkey_checksum
from silkaj.wot.tools import wot_lookup


def display_identity(idty: Identity) -> Texttable:
    """
    Creates a table containing the identity infos
    """
    client = client_instance()
    id_table = []
    id_table.append(["Public key", gen_pubkey_checksum(idty.pubkey)])
    id_table.append(["User ID", idty.uid])
    id_table.append(["Blockstamp", str(idty.block_id)])
    creation_block = client(bma.blockchain.block, idty.block_id.number)
    creation_date = pendulum.from_timestamp(creation_block["time"], tz="local").format(
        ALL
    )
    id_table.append(["Created on", creation_date])
    # display infos
    table = Texttable(max_width=shutil.get_terminal_size().columns)
    table.add_rows(id_table, header=False)
    return table


def check_many_identities(document: Union[Identity, Revocation]) -> bool:
    """
    Checks if many identities match the one looked after.
    Returns True if the same identity is found, False if not.
    """
    doc_type = document.__class__.__name__
    error_no_identical_id = f"{doc_type} document does not match any valid identity."
    idty = document if doc_type == "Identity" else document.identity

    try:
        results_pubkey = wot_lookup(idty.pubkey)
        results_uid = wot_lookup(idty.uid)
    except urllib.error.HTTPError:
        sys.exit(
            f"{error_no_identical_id}\nuid: {idty.uid}\npubkey: \
{gen_pubkey_checksum(idty.pubkey)}"
        )

    # get all matching identities
    lookup_ids = merge_ids_lists(results_pubkey, results_uid, idty.currency)
    match = False
    for n, lookup in enumerate(lookup_ids):
        if idty == lookup:
            lookup_ids.pop(n)
            match = True
            break
    alternate_ids = display_alternate_ids(lookup_ids).draw()
    if match:
        if len(lookup_ids) >= 1:
            click.echo(f"One matching identity!\nSimilar identities:\n{alternate_ids}")
        return True
    click.echo(f"{error_no_identical_id}\nSimilar identities:\n{alternate_ids}")
    return False


def display_alternate_ids(ids_list: List) -> Texttable:
    labels = ["uid", "public key", "timestamp"]
    table = Texttable(max_width=shutil.get_terminal_size().columns)
    table.header(labels)
    for _id in ids_list:
        table.add_row(
            [_id.uid, gen_pubkey_checksum(_id.pubkey), str(_id.block_id)[:12]]
        )
    return table


def merge_ids_lists(lookups_pubkey: List, lookups_uid: List, currency: str) -> List:
    """
    merge two lists of identities and remove duplicate identities.
    """
    ids = ids_list_from_lookups(lookups_pubkey, currency)
    ids_uid = ids_list_from_lookups(lookups_uid, currency)
    for _id in ids_uid:
        # __equal__ does not work. This is condition "id in ids".
        for listed_id in ids:
            if _id.signed_raw() == listed_id.signed_raw():
                id_in_ids = True
                break
            id_in_ids = False
        if not id_in_ids:
            ids.append(_id)
    return ids


def ids_list_from_lookups(lookups: List, currency: str) -> List:
    ids = []
    for lookup in lookups:
        pubkey = lookup["pubkey"]
        lookup_ids = lookup["uids"]
        for _id in lookup_ids:
            appended_id = Identity(
                currency=currency,
                pubkey=pubkey,
                uid=_id["uid"],
                block_id=BlockID.from_str(_id["meta"]["timestamp"]),
            )
            appended_id.signature = _id["self"]
            ids.append(appended_id)
    return ids



# Copyright  2016-2022 Maël Azimi <m.a@moul.re>
#
# Silkaj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Silkaj is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Silkaj. If not, see <https://www.gnu.org/licenses/>.

import functools
import re
import sys
from typing import Any
from urllib.error import HTTPError

from duniterpy import constants as du_const
from duniterpy.api import endpoint as ep
from duniterpy.api.client import Client
from duniterpy.documents import Document

from silkaj import constants


def determine_endpoint() -> ep.Endpoint:
    """
    Pass custom endpoint, parse through a regex
    {host|ipv4|[ipv6]}:{port}{/path}
    ^(?:(HOST)|(IPV4|[(IPV6)]))(?::(PORT))?(?:/(PATH))?$
    If gtest flag passed, return default gtest endpoint
    Else, return g1 default endpoint
    """

    regex = f"^(?:(?P<host>{du_const.HOST_REGEX})|(?P<ipv4>{du_const.IPV4_REGEX})|\
(?:\\[(?P<ipv6>{du_const.IPV6_REGEX})\\]))(?::(?P<port>{du_const.PORT_REGEX}))?\
(?:/(?P<path>{du_const.PATH_REGEX}))?$"

    try:
        from click.globals import get_current_context

        ctx = get_current_context()
        endpoint = ctx.obj["ENDPOINT"]
        gtest = ctx.obj["GTEST"]
    except (ModuleNotFoundError, RuntimeError):
        endpoint, gtest = None, None

    if endpoint:
        m = re.search(re.compile(regex), endpoint)
        if not m:
            sys.exit(
                "Error: Passed endpoint is of wrong format.\n\
Expected format: {host|ipv4|[ipv6]}:{port}{/path}"
            )
        port = int(m["port"]) if m["port"] else 443
        host, ipv4 = ep.fix_host_ipv4_mix_up(m["host"], m["ipv4"])

        if port == 443:
            return ep.SecuredBMAEndpoint(host, ipv4, m["ipv6"], port, m["path"])
        return ep.BMAEndpoint(host, ipv4, m["ipv6"], port)

    if gtest:
        return ep.endpoint(constants.G1_TEST_DEFAULT_ENDPOINT)
    return ep.endpoint(constants.G1_DEFAULT_ENDPOINT)


@functools.lru_cache(maxsize=1)
def client_instance():
    return Client(determine_endpoint())


def send_document(bma_path: Any, document: Document) -> None:
    client = client_instance()
    doc_name = document.__class__.__name__
    try:
        client(bma_path, document.signed_raw())
        print(f"{doc_name} successfully sent")
    except HTTPError as error:
        print(error)
        sys.exit(f"Error while publishing {doc_name.lower()}")


def exit_on_http_error(error: HTTPError, err_code: int, message: str) -> None:
    """
    Nicely displays a message on an expected error code.
    Else, displays the HTTP error message.
    """
    if error.code == err_code:
        sys.exit(message)
    print(error)
    sys.exit(constants.FAILURE_EXIT_STATUS)



# Copyright  2016-2022 Maël Azimi <m.a@moul.re>
#
# Silkaj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Silkaj is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Silkaj. If not, see <https://www.gnu.org/licenses/>.

import sys

from click import Context, group, help_option, option, pass_context, version_option
from duniterpy.api.endpoint import endpoint as du_endpoint

from silkaj.about import about
from silkaj.auth import generate_auth_file
from silkaj.blockchain.blocks import list_blocks
from silkaj.blockchain.difficulty import difficulties
from silkaj.blockchain.information import currency_info
from silkaj.blockchain.verify import verify_blocks_signatures
from silkaj.checksum import checksum_command
from silkaj.constants import (
    G1_DEFAULT_ENDPOINT,
    G1_TEST_DEFAULT_ENDPOINT,
    SILKAJ_VERSION,
)
from silkaj.g1_monetary_license import license_command
from silkaj.money.balance import balance_cmd
from silkaj.money.history import transaction_history
from silkaj.money.transfer import transfer_money
from silkaj.wot import revocation
from silkaj.wot.certification import send_certification
from silkaj.wot.lookup import lookup_cmd
from silkaj.wot.membership import send_membership
from silkaj.wot.status import status


@group()
@help_option("-h", "--help")
@version_option(SILKAJ_VERSION, "-v", "--version")
@option(
    "--endpoint",
    "-ep",
    help=f"Default endpoint to reach Ğ1 currency by its official node\
 {du_endpoint(G1_DEFAULT_ENDPOINT).host}\
 This option allows to specify a custom endpoint as follow: <host>:<port>.\
 In case no port is specified, it defaults to 443.",
)
@option(
    "--gtest",
    "-gt",
    is_flag=True,
    help=f"Default endpoint to reach ĞTest currency by its official node: \
{du_endpoint(G1_TEST_DEFAULT_ENDPOINT).host}",
)
@option(
    "--auth-scrypt",
    "--scrypt",
    is_flag=True,
    help="Scrypt authentication: default method",
)
@option("--nrp", help='Scrypt parameters: defaults N,r,p: "4096,16,1"')
@option(
    "--auth-file",
    "-af",
    is_flag=True,
    help="Authentication file. Defaults to: './authfile'",
)
@option(
    "--file",
    default="authfile",
    show_default=True,
    help="Path file specification with '--auth-file'",
)
@option("--auth-seed", "--seed", is_flag=True, help="Seed hexadecimal authentication")
@option("--auth-wif", "--wif", is_flag=True, help="WIF and EWIF authentication methods")
@option(
    "--display",
    "-d",
    is_flag=True,
    help="Display the generated document before sending it",
)
@option(
    "--dry-run",
    "-n",
    is_flag=True,
    help="By-pass licence, confirmation. \
Do not send the document, but display it instead",
)
@pass_context
def cli(
    ctx: Context,
    endpoint: str,
    gtest: bool,
    auth_scrypt: bool,
    nrp: str,
    auth_file: bool,
    file: str,
    auth_seed: bool,
    auth_wif: bool,
    display: bool,
    dry_run: bool,
) -> None:
    if display and dry_run:
        sys.exit("ERROR: display and dry-run options can not be used together")

    ctx.obj = {}
    ctx.ensure_object(dict)
    ctx.obj["ENDPOINT"] = endpoint
    ctx.obj["GTEST"] = gtest
    ctx.obj["AUTH_SCRYPT"] = auth_scrypt
    ctx.obj["AUTH_SCRYPT_PARAMS"] = nrp
    ctx.obj["AUTH_FILE"] = auth_file
    ctx.obj["AUTH_FILE_PATH"] = file
    ctx.obj["AUTH_SEED"] = auth_seed
    ctx.obj["AUTH_WIF"] = auth_wif
    ctx.obj["DISPLAY_DOCUMENT"] = display
    ctx.obj["DRY_RUN"] = dry_run


cli.add_command(about)
cli.add_command(generate_auth_file)
cli.add_command(checksum_command)
cli.add_command(license_command)


@cli.group("blockchain", help="Blockchain related commands")
@help_option("-h", "--help")
def blockchain_group() -> None:
    pass


blockchain_group.add_command(list_blocks)
blockchain_group.add_command(difficulties)
blockchain_group.add_command(currency_info)
blockchain_group.add_command(verify_blocks_signatures)


@cli.group("money", help="Money management related commands")
@help_option("-h", "--help")
def money_group() -> None:
    pass


money_group.add_command(balance_cmd)
money_group.add_command(transaction_history)
money_group.add_command(transfer_money)


@cli.group("wot", help="Web-of-Trust related commands")
@help_option("-h", "--help")
def wot_group() -> None:
    pass


wot_group.add_command(send_certification)
wot_group.add_command(lookup_cmd)
wot_group.add_command(send_membership)
wot_group.add_command(status)


@wot_group.group(
    "revocation",
    help="Create, save, verify or publish revocation document.\n\
Subcommands optionally take the path to the revocation document.",
)
@help_option("-h", "--help")
def revocation_group() -> None:
    pass


revocation_group.add_command(revocation.save)
revocation_group.add_command(revocation.verify)
revocation_group.add_command(revocation.publish)
revocation_group.add_command(revocation.revoke_now)



# Copyright  2016-2022 Maël Azimi <m.a@moul.re>
#
# Silkaj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Silkaj is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Silkaj. If not, see <https://www.gnu.org/licenses/>.

import sys
from typing import Dict

import click
from duniterpy.api import bma
from duniterpy.api.client import Client
from duniterpy.documents import Block, BlockID, Certification, Identity, get_block_id
from duniterpy.key import SigningKey
from pendulum import from_timestamp, now

from silkaj import tui
from silkaj.auth import auth_method
from silkaj.blockchain import tools as bc_tools
from silkaj.constants import ALL, DATE
from silkaj.g1_monetary_license import license_approval
from silkaj.network import client_instance, send_document
from silkaj.public_key import gen_pubkey_checksum, is_pubkey_and_check
from silkaj.wot import tools as wot_tools


@click.command("certification", help="Send certification")
@click.argument("uid_pubkey_to_certify")
@click.pass_context
def send_certification(ctx: click.Context, uid_pubkey_to_certify: str) -> None:
    client = client_instance()

    checked_pubkey = is_pubkey_and_check(uid_pubkey_to_certify)
    if checked_pubkey:
        uid_pubkey_to_certify = str(checked_pubkey)

    # pylint: disable=unused-variable
    idty_to_certify, pubkey_to_certify, send_certs = wot_tools.choose_identity(
        uid_pubkey_to_certify
    )

    # Authentication
    key = auth_method()

    issuer_pubkey = key.pubkey
    issuer = pre_checks(client, issuer_pubkey, pubkey_to_certify)

    # Display license and ask for confirmation
    head = bc_tools.get_head_block()
    currency = head["currency"]
    license_approval(currency)

    # Certification confirmation
    certification_confirmation(
        ctx, issuer, issuer_pubkey, pubkey_to_certify, idty_to_certify
    )

    # Create and sign certification document
    certification = docs_generation(
        currency,
        pubkey_to_certify,
        idty_to_certify,
        issuer_pubkey,
        head,
        key,
    )

    if ctx.obj["DISPLAY_DOCUMENT"]:
        click.echo(certification.signed_raw(), nl=False)
        tui.send_doc_confirmation("certification")

    # Send certification document
    send_document(bma.wot.certify, certification)


def pre_checks(client: Client, issuer_pubkey: str, pubkey_to_certify: str) -> Dict:
    # Check whether current user is member
    issuer = wot_tools.is_member(issuer_pubkey)
    if not issuer:
        sys.exit("Current identity is not member.")

    if issuer_pubkey == pubkey_to_certify:
        sys.exit("You can’t certify yourself!")

    # Check if the certification can be renewed
    params = bc_tools.get_blockchain_parameters()
    requirements = client(bma.wot.requirements, pubkey_to_certify)
    req = requirements["identities"][0]  # type: Dict
    for cert in req["certifications"]:
        if cert["from"] == issuer_pubkey:
            # Ğ1: 0<–>2y - 2y + 2m
            # ĞT: 0<–>4.8m - 4.8m + 12.5d
            renewable = cert["expiresIn"] - params["sigValidity"] + params["sigReplay"]
            if renewable > 0:
                renewable_date = now().add(seconds=renewable).format(DATE)
                sys.exit(f"Certification renewable from {renewable_date}")

    # Check if the certification is already in the pending certifications
    for pending_cert in req["pendingCerts"]:
        if pending_cert["from"] == issuer_pubkey:
            sys.exit("Certification is currently being processed")
    return issuer


def certification_confirmation(
    ctx: click.Context,
    issuer: Dict,
    issuer_pubkey: str,
    pubkey_to_certify: str,
    idty_to_certify: Dict,
) -> None:
    cert = []
    client = client_instance()
    idty_timestamp = idty_to_certify["meta"]["timestamp"]
    block_id_idty = get_block_id(idty_timestamp)
    block = client(bma.blockchain.block, block_id_idty.number)
    timestamp_date = from_timestamp(block["time"], tz="local").format(ALL)
    block_id_date = f": #{idty_timestamp[:15]}… {timestamp_date}"
    cert.append(["ID", issuer["uid"], "–>", idty_to_certify["uid"] + block_id_date])
    cert.append(
        [
            "Pubkey",
            gen_pubkey_checksum(issuer_pubkey),
            "–>",
            gen_pubkey_checksum(pubkey_to_certify),
        ]
    )
    params = bc_tools.get_blockchain_parameters()
    cert_ends = now().add(seconds=params["sigValidity"]).format(DATE)
    cert.append(["Valid", now().format(DATE), "—>", cert_ends])

    table = tui.Table()
    table.fill_rows(
        cert, ["Cert", "Issuer", "–>", "Recipient: Published: #block-hash date"]
    )
    click.echo(table.draw())

    if not ctx.obj["DISPLAY_DOCUMENT"]:
        tui.send_doc_confirmation("certification")


def docs_generation(
    currency: str,
    pubkey_to_certify: str,
    idty_to_certify: Dict,
    issuer_pubkey: str,
    head: Block,
    key: SigningKey,
) -> Certification:
    identity = Identity(
        pubkey=pubkey_to_certify,
        uid=idty_to_certify["uid"],
        block_id=get_block_id(idty_to_certify["meta"]["timestamp"]),
        currency=currency,
    )
    identity.signature = idty_to_certify["self"]

    return Certification(
        pubkey_from=issuer_pubkey,
        identity=identity,
        block_id=BlockID(head["number"], head["hash"]),
        signing_key=key,
        currency=currency,
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['silkaj', 'silkaj.blockchain', 'silkaj.money', 'silkaj.wot']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.0.0,<9.0.0',
 'duniterpy==1.1.0',
 'pendulum>=2.1.2,<3.0.0',
 'texttable>=1.6.3,<2.0.0']

entry_points = \
{'console_scripts': ['silkaj = silkaj.cli:cli']}

setup_kwargs = {
    'name': 'silkaj',
    'version': '0.11.0rc0',
    'description': 'Powerfull, lightweight, and multi-platform command line client written with Python for Duniter’s currencies: Ğ1 and Ğ1-Test.',
    'long_description': '<img src="https://git.duniter.org/clients/python/silkaj/raw/main/logo/silkaj_logo.svg" width="250" />\n\n# Silkaj\n\n[![Version](https://img.shields.io/pypi/v/silkaj.svg)](https://pypi.python.org/pypi/silkaj)\n[![License](https://img.shields.io/pypi/l/silkaj.svg)](https://pypi.python.org/pypi/silkaj)\n[![Python versions](https://img.shields.io/pypi/pyversions/silkaj.svg?logo=python&label=Python&logoColor=gold)](https://pypi.python.org/pypi/silkaj)\n[![Code format](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![Imports: isort](https://img.shields.io/badge/%20imports-isort-%231674b1?style=flat&labelColor=ef8336)](https://pycqa.github.io/isort/)\n[![Coverage report](https://git.duniter.org/clients/python/silkaj/badges/main/coverage.svg)](https://clients.duniter.io/python/silkaj/index.html)\n[![Website](https://img.shields.io/website/https/silkaj.duniter.org.svg)](https://silkaj.duniter.org)\n[![Dev pipeline status](https://git.duniter.org/clients/python/silkaj/badges/main/pipeline.svg)](https://git.duniter.org/clients/python/silkaj/)\n[![types - Mypy](https://img.shields.io/badge/types-Mypy-blue.svg)](http://www.mypy-lang.org/)\n\nPowerfull, lightweight, and multi-platform command line client written with Python for Ğ1 and Ğ1-Test currencies\n\n- [Website](https://silkaj.duniter.org)\n\n## Install\n\n### Distribution\n\nInstall with your favorite package manager. See below the [packaging status paragraph](#packaging-status).\n\n### Pipx\n\nIf you want a more recent version [install with pipx](doc/install_pipx.md):\n\n```bash\nsudo apt install pipx\npipx install silkaj\n```\n\n### Docker images\n\nThere is two kind of images. One build with `pip` for user purposes, and one using Poetry for developer purposes.\n\n- [Docker images](doc/docker.md)\n\n### For contributing purposes\n\n- [Install the Poetry development environment](doc/install_poetry.md)\n- Check out the [contributing guidelines](CONTRIBUTING.md)\n\n## Usage\n\n- Get help usage with `-h` or `--help` options, then run:\n\n```bash\nsilkaj <sub-command>\n```\n\n- Will automatically request and post data on `duniter.org 443` main Ğ1 node.\n\n- Specify a custom node with `-ep` option:\n\n```bash\nsilkaj -ep <hostname>:<port> <sub-command>\n```\n\n## Features\n\n### Currency information & blockchain exploration\n\n- Check the present currency information stand\n- Display current proof of work difficulty level to generate the next block\n- Explore the blockchain block by block\n- Verify blockchain blocks hashes\n\n### Money management\n\n- Transaction emission\n  - Multi-recipients transaction support\n  - Read transaction recipients and amounts from a file\n- Consult wallets balances\n- Consult wallet history\n\n### Web-of-Trust management\n\n- Look up for public keys and identities\n- Check sent and received certifications and consult the membership status of any given identity in the Web of Trust\n- Certification emission\n- Membership emission\n- Revocation file handling\n\n### Authentication\n\n- Authentication methods: Scrypt, file, and (E)WIF\n\n### Others\n\n- Display Ğ1 monetary license\n- Public key checksum\n\n## Wrappers\n\n- [How-to: automate transactions and multi-output](doc/how-to_automate_transactions_and_multi-output.md)\n- [Transaction generator written in Shell](https://gitlab.com/jytou/tgen)\n- [Ğ1Cotis](https://git.duniter.org/matograine/g1-cotis)\n- [G1pourboire](https://git.duniter.org/matograine/g1pourboire)\n- [Ğ1SMS](https://git.duniter.org/clients/G1SMS/)\n- [Ğmixer](https://git.duniter.org/tuxmain/gmixer-py/)\n- [printqrjune](https://github.com/jbar/printqrjune)\n\n### Dependencies\n\nSilkaj is based on following Python modules:\n\n- [Click](https://click.palletsprojects.com/): Composable command line interface toolkit\n- [DuniterPy](https://git.duniter.org/clients/python/duniterpy/): Most complete client oriented Python library for Duniter/Ğ1 ecosystem\n- [Pendulum](https://pendulum.eustace.io/): Datetimes made easy\n- [texttable](https://github.com/foutaise/texttable/): Creation of simple ASCII tables\n\n### Names\n\nI wanted to call that program:\n\n- bamiyan\n- margouillat\n- lsociety\n- cashmere\n\nI finally called it `Silkaj` as `Silk` in esperanto.\n\n### Website\n\n- [Silkaj website sources](https://git.duniter.org/websites/silkaj_website/)\n\n## Packaging status\n\n[![Packaging status](https://repology.org/badge/vertical-allrepos/silkaj.svg)](https://repology.org/project/silkaj/versions)\n',
    'author': 'Moul',
    'author_email': 'moul@moul.re',
    'maintainer': 'Moul',
    'maintainer_email': 'moul@moul.re',
    'url': 'https://silkaj.duniter.org',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7.0,<4.0.0',
}


setup(**setup_kwargs)



# Copyright  2016-2022 Maël Azimi <m.a@moul.re>
#
# Silkaj is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Silkaj is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with Silkaj. If not, see <https://www.gnu.org/licenses/>.

import sys
from typing import List

from click import Context, argument, command, echo, pass_context

from silkaj import tui
from silkaj.auth import auth_method, has_auth_method
from silkaj.blockchain.tools import get_head_block
from silkaj.money import tools as m_tools
from silkaj.public_key import gen_pubkey_checksum, is_pubkey_and_check
from silkaj.tools import get_currency_symbol
from silkaj.wot import tools as wt


@command("balance", help="Get wallet balance")
@argument("pubkeys", nargs=-1)
@pass_context
def balance_cmd(ctx: Context, pubkeys: str) -> None:
    if not has_auth_method():

        # check input pubkeys
        if not pubkeys:
            sys.exit("You should specify one or many pubkeys")
        pubkeys_list = []
        wrong_pubkeys = False
        for input_pubkey in pubkeys:
            checked_pubkey = is_pubkey_and_check(input_pubkey)
            if checked_pubkey:
                pubkey = str(checked_pubkey)
            else:
                pubkey = input_pubkey
                wrong_pubkeys = True
                print(f"ERROR: pubkey {pubkey} has a wrong format")
            if pubkey in pubkeys_list:
                sys.exit(
                    f"ERROR: pubkey {gen_pubkey_checksum(pubkey)} was specified many times"
                )
            pubkeys_list.append(pubkey)
        if wrong_pubkeys:
            sys.exit("Please check the pubkeys format.")

        total = [0, 0]
        for pubkey in pubkeys_list:
            inputs_balance = m_tools.get_amount_from_pubkey(pubkey)
            show_amount_from_pubkey(pubkey, inputs_balance)
            total[0] += inputs_balance[0]
            total[1] += inputs_balance[1]
        if len(pubkeys_list) > 1:
            show_amount_from_pubkey("Total", total)
    else:
        key = auth_method()
        pubkey = key.pubkey
        show_amount_from_pubkey(pubkey, m_tools.get_amount_from_pubkey(pubkey))


def show_amount_from_pubkey(label: str, inputs_balance: List[int]) -> None:
    """
    Shows the balance of a pubkey.
    `label` can be either a pubkey or "Total".
    """
    totalAmountInput = inputs_balance[0]
    balance = inputs_balance[1]
    currency_symbol = get_currency_symbol()
    ud_value = m_tools.get_ud_value()
    average = get_average()
    member = None

    # if `pubkey` is a pubkey, get pubkey:checksum and uid
    if label != "Total":
        member = wt.is_member(label)
        label = gen_pubkey_checksum(label)
    # display balance table
    display = []
    display.append(["Balance of pubkey", label])

    if member:
        display.append(["User identifier", member["uid"]])

    if totalAmountInput - balance != 0:
        m_tools.display_amount(
            display, "Blockchain", balance, ud_value, currency_symbol
        )
        m_tools.display_amount(
            display,
            "Pending transaction",
            (totalAmountInput - balance),
            ud_value,
            currency_symbol,
        )
    m_tools.display_amount(
        display, "Total balance", totalAmountInput, ud_value, currency_symbol
    )
    display.append(
        [
            "Total relative to M/N",
            f"{round(totalAmountInput / average, 2)} x M/N",
        ]
    )

    table = tui.Table()
    table.fill_rows(display)
    echo(table.draw())


def get_average() -> int:
    head = get_head_block()
    return head["monetaryMass"] / head["membersCount"]
